'use strict';

var tslib_1 = require('tslib');
var inversify = require('inversify');
var path = require('path');
var fs = require('fs');
var child_process = require('child_process');
var fbWatchman = require('fb-watchman');
require('reflect-metadata');

var Bindings = {
    Cli: Symbol('Cli'),
    Config: Symbol('Config'),
    ConfigManager: Symbol('ConfigManager'),
    Process: Symbol('Process'),
    Require: Symbol('require'),
    Spawn: Symbol('spawn'),
    Sync: Symbol('Sync'),
    Terminal: Symbol('Terminal'),
    WatchmanClient: Symbol('WatchmanClient'),
    WatchmanProcessor: Symbol('WatchmanProcessor'),
};

var DEFAULT_ARGUMENTS = {
    config: '',
    init: false,
};
var CliImpl = /** @class */ (function () {
    function CliImpl(process) {
        this.process = process;
    }
    CliImpl.prototype.getArguments = function () {
        var userArgs = Object.assign({}, DEFAULT_ARGUMENTS);
        var args = this.process.argv;
        for (var i = 0; i < args.length; i++) {
            var arg = args[i];
            switch (arg) {
                case '-i':
                case '--init':
                    userArgs.init = true;
                    break;
                case '-c':
                case '--config':
                    userArgs.config = this.parseConfigFile(i, args);
                    break;
                default:
                    break;
            }
        }
        return userArgs;
    };
    CliImpl.prototype.parseConfigFile = function (i, args) {
        if (i + 1 < args.length) {
            return path.resolve(args[i + 1]);
        }
        return this.process.cwd() + '/.watchman-processor.config.js';
    };
    CliImpl = tslib_1.__decorate([
        inversify.injectable(),
        tslib_1.__param(0, inversify.inject(Bindings.Process))
    ], CliImpl);
    return CliImpl;
}());

var INIT_MSG = 'The watchman-processor configuration does not exist. \n\n' +
    'Run "watchman-processor --init" to create an example configuration file.\n';
var DEFAULT_CONFIG = {
    controlWatchman: true,
    debug: false,
    emoji: true,
    maxFileLength: 100,
    rsyncCmd: 'rsync',
    shell: '/bin/sh',
};
var DEFAULT_SUBSCRIPTION = {
    ignoreFolders: [],
    type: 'rsync',
};
var ConfigManagerImpl = /** @class */ (function () {
    function ConfigManagerImpl(cli, require, process) {
        this.cli = cli;
        this.require = require;
        this.process = process;
        /* istanbul ignore next */
        var homeFolder = this.process.env[(this.process.platform === 'win32') ? 'USERPROFILE' : 'HOME'];
        var args = this.cli.getArguments();
        this.options = Object.assign({
            confFile: args.config || path.resolve(homeFolder + '/.watchman-processor.config.js'),
            exampleConfFile: path.resolve(__dirname + '/example/watchman-processor.config.js'),
        });
    }
    ConfigManagerImpl.prototype.getConfig = function () {
        var _a = this, cachedConfig = _a.cachedConfig, options = _a.options, require = _a.require;
        if (cachedConfig) {
            return cachedConfig;
        }
        try {
            var configFile = path.resolve(options.confFile);
            if (!fs.existsSync(configFile)) {
                var error = new Error(INIT_MSG);
                error.name = 'init';
                throw error;
            }
            var userConfig = require(configFile);
            var config = this.cachedConfig = Object.assign({}, DEFAULT_CONFIG, userConfig);
            /* istanbul ignore if */
            if (!config.subscriptions) {
                throw new Error('No subscriptions defined');
            }
            // set the default settings for subscription configs
            var subscriptions = Object.keys(config.subscriptions);
            for (var _i = 0, subscriptions_1 = subscriptions; _i < subscriptions_1.length; _i++) {
                var name = subscriptions_1[_i];
                config.subscriptions[name] = Object.assign({}, DEFAULT_SUBSCRIPTION, config.subscriptions[name]);
            }
            return config;
        }
        catch (e) {
            return e;
        }
    };
    ConfigManagerImpl.prototype.createConfig = function (
    /* istanbul ignore next */
    confFile, 
    /* istanbul ignore next */
    exampleConfFile) {
        /* istanbul ignore next */
        if (confFile === void 0) { confFile = this.options.confFile; }
        /* istanbul ignore next */
        if (exampleConfFile === void 0) { exampleConfFile = this.options.exampleConfFile; }
        var _a = this, options = _a.options, process = _a.process;
        return new Promise(function (done, reject) {
            var reader = fs.createReadStream(exampleConfFile);
            var writer = fs.createWriteStream(confFile);
            reader.on('error', reject);
            writer.on('error', reject);
            writer.on('close', function () {
                process.stdout.write('Done.  "' + options.confFile + '" created.\n');
                done();
            });
            reader.pipe(writer);
        });
    };
    ConfigManagerImpl = tslib_1.__decorate([
        inversify.injectable(),
        tslib_1.__param(0, inversify.inject(Bindings.Cli)),
        tslib_1.__param(1, inversify.inject(Bindings.Require)),
        tslib_1.__param(2, inversify.inject(Bindings.Process))
    ], ConfigManagerImpl);
    return ConfigManagerImpl;
}());

var SyncImpl = /** @class */ (function () {
    function SyncImpl(config, terminal, spawn) {
        this.config = config;
        this.terminal = terminal;
        this.spawn = spawn;
        this.processes = new Set();
    }
    SyncImpl.prototype.syncFiles = function (subConfig, filesNames) {
        var ignoreFolders = subConfig.ignoreFolders;
        // remove files that are in the ignore folders
        var files = (filesNames || []).filter(function (file) { return !exists(ignoreFolders, file); });
        // if there are too many files,
        // it might just be better to let rsync figure out what needs to be synced
        if (files.length > 0 && files.length < this.config.maxFileLength) {
            return this._syncSpecificFiles(subConfig, files);
        }
        else {
            return this._syncAllFiles(subConfig);
        }
    };
    SyncImpl.prototype.end = function () {
        var _a = this, processes = _a.processes, terminal = _a.terminal;
        /* istanbul ignore else */
        if (processes.size > 0) {
            terminal.debug("sync: kill rsync");
            processes.forEach(function (proc) {
                proc.kill();
            });
        }
    };
    SyncImpl.prototype._syncAllFiles = function (subConfig) {
        var destination = subConfig.destination, ignoreFolders = subConfig.ignoreFolders, source = subConfig.source;
        var excludes = ("--exclude '" + ignoreFolders.join("' --exclude '") + "'").split(' ');
        var args = ['-avz', '--delete'].concat(excludes, [source, destination]);
        return this._exec(args);
    };
    SyncImpl.prototype._syncSpecificFiles = function (subConfig, files) {
        files = getUniqueFileFolders(files).concat(files);
        var destination = subConfig.destination, source = subConfig.source;
        var includes = ("--include '" + files.join("' --include '") + "'").split(' ');
        var args = ['-avz', '--delete'].concat(includes, ['--exclude', "'*'", source, destination]);
        return this._exec(args);
    };
    SyncImpl.prototype._exec = function (args) {
        var _this = this;
        var _a = this, spawn = _a.spawn, config = _a.config, terminal = _a.terminal;
        var cmdAndArgs = config.rsyncCmd + ' ' + args.join(' ');
        terminal.debug(cmdAndArgs);
        return new Promise(function (resolve, reject) {
            var child = spawn(config.shell, ['-c', cmdAndArgs]);
            var errBuffer = [];
            _this.processes.add(child);
            var done = _this._execDoneHandler(errBuffer, child, resolve, reject);
            child.on('exit', done);
            child.on('close', done);
            child.stderr.on('data', _this._execErrDataHandler(errBuffer, terminal));
            child.stdout.on('data', _this._execDataHandler(terminal));
            child.on('error', _this._execErrorHandler(child, reject));
        });
    };
    SyncImpl.prototype._execDataHandler = function (terminal) {
        return function (data) {
            terminal.debug(toString(data));
        };
    };
    SyncImpl.prototype._execErrDataHandler = function (errBuffer, terminal) {
        return function (data) {
            data = toString(data);
            errBuffer.push(data);
            terminal.debug(data);
        };
    };
    SyncImpl.prototype._execDoneHandler = function (errBuffer, child, resolve, reject) {
        var _this = this;
        return function (code) {
            _this.processes.delete(child);
            code > 0 ? reject(errBuffer.join('\n')) : resolve();
        };
    };
    SyncImpl.prototype._execErrorHandler = function (child, reject) {
        var _this = this;
        return function (err) {
            _this.processes.delete(child);
            reject(err);
        };
    };
    SyncImpl = tslib_1.__decorate([
        inversify.injectable(),
        tslib_1.__param(0, inversify.inject(Bindings.Config)),
        tslib_1.__param(1, inversify.inject(Bindings.Terminal)),
        tslib_1.__param(2, inversify.inject(Bindings.Spawn))
    ], SyncImpl);
    return SyncImpl;
}());
/**
 * Convert Buffer to string
 */
function toString(data) {
    /* istanbul ignore next */
    return data instanceof Buffer ? data.toString('utf8') : data;
}
/**
 * Given a list of files, return a list of folders leading up to those file names.
 *  Example:
 *    Given: ['example/js/1.js', 'example/js/2/.js']
 *    Will Return: ['example/', 'example/js/', 'example/js/1.js', 'example/js/2/.js']
 */
function getUniqueFileFolders(files) {
    var folders = [];
    var seen = new Map();
    for (var _i = 0, files_1 = files; _i < files_1.length; _i++) {
        var file = files_1[_i];
        var folderParts = getFolderParts(file.split('/'), seen);
        folders.push.apply(folders, folderParts);
    }
    return folders;
}
function getFolderParts(folderParts, seen) {
    var folders = [];
    var folderPartsSum = '';
    for (var _i = 0, folderParts_1 = folderParts; _i < folderParts_1.length; _i++) {
        var folderPart = folderParts_1[_i];
        if (folderPartsSum.length > 0) {
            folderPartsSum += '/';
        }
        folderPartsSum += folderPart;
        if (!seen.has(folderPartsSum)) {
            seen.set(folderPartsSum, undefined);
            folders.push(folderPartsSum);
        }
    }
    return folders;
}
/*
 * Return whether a value starts with a searchString in the given values array
 */
function exists(values, stringToSearch) {
    for (var _i = 0, values_1 = values; _i < values_1.length; _i++) {
        var value = values_1[_i];
        if (startsWith(stringToSearch, value)) {
            return true;
        }
    }
    return false;
}
/**
 * Return whether a string starts with another string
 */
function startsWith(value, searchString) {
    return value.substr(0, searchString.length) === searchString;
}

var fgBlack = '\x1b[30m';
var bgRed = '\x1b[41m';
var bgGreen = '\x1b[42m';
var bgYellow = '\x1b[43m';
var bgWhite = '\x1b[47m';
var TerminalImpl = /** @class */ (function () {
    function TerminalImpl(config, process) {
        this.config = config;
        this.process = process;
    }
    TerminalImpl.prototype.error = function (err) {
        var msg = err.toString();
        this.process.stderr.write("" + msg);
        if (typeof err !== 'string') {
            this.process.stderr.write("" + err.stack);
        }
    };
    TerminalImpl.prototype.debug = function (msg) {
        if (this.config.debug && typeof msg === 'string' && msg.length > 0) {
            msg = msg.trim();
            this.process.stdout.write(msg + '\n');
        }
    };
    TerminalImpl.prototype.setState = function (configEntry, state, statusMessage) {
        configEntry.state = state;
        configEntry.statusMessage = statusMessage;
        this.render();
    };
    TerminalImpl.prototype.render = function () {
        /* istanbul ignore if */
        if (this._quiet()) {
            return;
        }
        this._clear();
        var subscriptions = Object.keys(this.config.subscriptions);
        var statusBuffer = [];
        for (var _i = 0, subscriptions_1 = subscriptions; _i < subscriptions_1.length; _i++) {
            var name = subscriptions_1[_i];
            var subscription = this.config.subscriptions[name];
            var state = subscription.state, statusMessage = subscription.statusMessage;
            if (statusMessage) {
                statusBuffer.push(statusMessage);
            }
            this._renderState(name, state);
        }
        this._log('\n', bgWhite);
        this._reset();
        this.error(statusBuffer.join('\n'));
    };
    TerminalImpl.prototype._renderState = function (name, state) {
        var log = this._log.bind(this);
        var emoji = this._emoji.bind(this);
        if (state === 'good') {
            log(emoji('👍 ') + ' ' + name + ' ', bgGreen);
        }
        else if (state === 'running') {
            log(emoji('🏃 ') + ' ' + name + ' ', bgYellow);
        }
        else if (state === 'error') {
            log(emoji('💀 ') + ' ' + name + ' ', bgRed);
        }
        else {
            log(emoji('⚡️ ') + ' ' + name + ' ', bgWhite);
        }
    };
    TerminalImpl.prototype._emoji = function (str) {
        /* istanbul ignore next */
        return this.config.emoji ? str : '';
    };
    TerminalImpl.prototype._clear = function () {
        this.process.stdout.write('\x1b[H\x1b[J');
    };
    TerminalImpl.prototype._reset = function () {
        this.process.stdout.write('\x1b[0m');
    };
    TerminalImpl.prototype._log = function (msg, backgroundColor) {
        this.process.stdout.write("" + backgroundColor + fgBlack + msg);
        this._reset();
    };
    TerminalImpl.prototype._quiet = function () {
        return (!this
            || !this.config
            || this.config.debug
            || !this.config.subscriptions);
    };
    TerminalImpl = tslib_1.__decorate([
        inversify.injectable(),
        tslib_1.__param(0, inversify.inject(Bindings.Config)),
        tslib_1.__param(1, inversify.inject(Bindings.Process))
    ], TerminalImpl);
    return TerminalImpl;
}());

var WatchmanProcessorImpl = /** @class */ (function () {
    function WatchmanProcessorImpl(config, client, terminal, sync) {
        this.config = config;
        this.client = client;
        this.terminal = terminal;
        this.sync = sync;
    }
    WatchmanProcessorImpl.prototype.start = function () {
        var _a = this, client = _a.client, terminal = _a.terminal;
        terminal.debug('watchman: initialize');
        var onCapabilityCheck = this.onCapabilityCheck.bind(this);
        client.capabilityCheck({}, onCapabilityCheck);
    };
    WatchmanProcessorImpl.prototype.end = function () {
        var _this = this;
        var _a = this, client = _a.client, config = _a.config, sync = _a.sync;
        var subscriptions = config.subscriptions;
        var promises = [];
        for (var _i = 0, _b = Object.keys(subscriptions); _i < _b.length; _i++) {
            var name = _b[_i];
            var sub = subscriptions[name];
            promises.push(this.unsubscribe(sub.source, name));
        }
        sync.end();
        return Promise.all(promises)
            .then(function () {
            if (config.controlWatchman) {
                _this.shutdown();
            }
            else {
                client.end();
            }
        });
    };
    WatchmanProcessorImpl.prototype.onCapabilityCheck = function (error) {
        var terminal = this.terminal;
        if (error) {
            terminal.error(error);
            return;
        }
        terminal.render();
        var client = this.client;
        var onSubscription = this.onSubscription.bind(this);
        var promises = [];
        var subscriptions = Object.keys(this.config.subscriptions);
        for (var _i = 0, subscriptions_1 = subscriptions; _i < subscriptions_1.length; _i++) {
            var name = subscriptions_1[_i];
            var sub = this.config.subscriptions[name];
            var expression = sub.watchExpression || ['allof', ['type', 'f']];
            for (var _a = 0, _b = sub.ignoreFolders; _a < _b.length; _a++) {
                var folder = _b[_a];
                expression.push(['not', ['dirname', folder]]);
            }
            promises.push(this.subscribe(path.resolve(sub.source), name, expression));
        }
        var render = terminal.render.bind(terminal);
        var errHandler = terminal.error.bind(terminal);
        Promise.all(promises).then(render).catch(errHandler);
        // subscription is fired regardless of which subscriber fired it
        client.on('subscription', onSubscription);
    };
    WatchmanProcessorImpl.prototype.onSubscription = function (resp) {
        var config = this.config;
        var subscription = resp && resp.subscription;
        var files = resp.files;
        var subConfig = config.subscriptions[subscription];
        this.syncFiles(subConfig, files);
    };
    WatchmanProcessorImpl.prototype.syncFiles = function (subConfig, files) {
        var _a = this, sync = _a.sync, terminal = _a.terminal;
        terminal.setState(subConfig, 'running');
        var fileNames = (files || []).map(function (file) { return file.name; });
        sync.syncFiles(subConfig, fileNames)
            .then(function () {
            terminal.setState(subConfig, 'good');
        })
            .catch(function (err) {
            terminal.setState(subConfig, 'error', err);
        });
    };
    WatchmanProcessorImpl.prototype.subscribe = function (folder, name, expression) {
        var terminal = this.terminal;
        var client = this.client;
        var sub = {
            expression: expression,
            fields: ['name', 'exists'],
            relative_root: '',
        };
        terminal.debug("subscribe: " + name);
        return new Promise(function (resolve, reject) {
            client.command(['subscribe', folder, name, sub], function (error) {
                error ? reject('failed to start: ' + error) : resolve();
            });
        });
    };
    WatchmanProcessorImpl.prototype.unsubscribe = function (folder, name) {
        var terminal = this.terminal;
        var client = this.client;
        terminal.debug("unsubscribe: " + name);
        return new Promise(function (resolve) {
            client.command(['unsubscribe', folder, name], resolve);
        });
    };
    WatchmanProcessorImpl.prototype.shutdown = function () {
        var terminal = this.terminal;
        var client = this.client;
        terminal.debug("watchman: shutdown");
        return new Promise(function (resolve) {
            client.command(['shutdown-server'], resolve);
        });
    };
    WatchmanProcessorImpl = tslib_1.__decorate([
        inversify.injectable(),
        tslib_1.__param(0, inversify.inject(Bindings.Config)),
        tslib_1.__param(1, inversify.inject(Bindings.WatchmanClient)),
        tslib_1.__param(2, inversify.inject(Bindings.Terminal)),
        tslib_1.__param(3, inversify.inject(Bindings.Sync))
    ], WatchmanProcessorImpl);
    return WatchmanProcessorImpl;
}());

var container = new inversify.Container();
// setup core functionality and externals
container.bind(Bindings.Process).toConstantValue(process);
container.bind(Bindings.Spawn).toConstantValue(child_process.spawn);
container.bind(Bindings.Require).toConstantValue(require);
container.bind(Bindings.WatchmanClient).toConstantValue(new fbWatchman.Client());
// setup the main classes
container.bind(Bindings.Cli).to(CliImpl);
container.bind(Bindings.ConfigManager).to(ConfigManagerImpl);
container.bind(Bindings.Config).toDynamicValue(function () {
    return container.get(Bindings.ConfigManager).getConfig();
});
container.bind(Bindings.Terminal).to(TerminalImpl);
container.bind(Bindings.Sync).to(SyncImpl);
container.bind(Bindings.WatchmanProcessor).to(WatchmanProcessorImpl);

var cli = container.get(Bindings.Cli);
var configManager = container.get(Bindings.ConfigManager);
var watchmanProcessor = container.get(Bindings.WatchmanProcessor);
var args = cli.getArguments();
var processor = watchmanProcessor;
if (args.init) {
    configManager.createConfig();
    process.abort();
}
else {
    var config = configManager.getConfig();
    if (config instanceof Error) {
        processor = null;
        process.stderr.write(config.message + '\n');
        if (config.name !== 'init') {
            process.stderr.write(config.name);
            if (config.stack) {
                process.stderr.write(config.stack);
            }
        }
    }
}
var processor$1 = processor;

module.exports = processor$1;
//# sourceMappingURL=index.js.map
